function [p,Plast,a,q] = LUbanded(A,r)

% Implements inversion algorithm for two-sided banded matrices
% from Theorem 5.6 in manuscript "Computation of
% quasiseparable representations of Green matrices".
% The algorithm computes lower Green generators of the
% inverse matrix.
%
% Input: A banded matrix to be inverted,
%        r semi-bandwidth
%          (i.e., a tridiagonal matrix has r = 1)

% Check size of A
[Nr,Nc] = size(A);
if Nr == Nc
    N = Nr;
else
    disp('Matrix is not square')
    return
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Step 1: compute generators of LR decomposition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Here R is upper triangular and upper banded
% with bandwidth r

R = zeros(N);
L = zeros(r+1,r+1,N-r+1);

% 1.1
gamma = A(1,1);
f = A(2:r+1,1)/gamma;
X = A(1,2:r+1);
Y = -f*[X, 0] + A(2:r+1,2:r+2);
L(:,:,1) = [1, zeros(1,r); -f, eye(r)];
R(1,1) = gamma;
R(1,2:r+1) = A(1,2:r+1);

% 1.2
for k = 2:N-r-1
   gamma = Y(1,1);
   X = Y(1,2:r+1);
   f = [Y(2:r,1); A(k+r,k)]/gamma;
   % Z = [Y(2:r+1,2:r+2); A(k+r,k+1:k+r+2)];  % non va
   % Z = [Y(2:r,2:r+1),A(k+r-1,k+r+1);... % per r=2, generalizzare
   %    A(k+r,k+1:k+r+1)];
   Z = [Y(2:r,2:r+1),A(k+1:k+r-1,k+r+1);... 
       A(k+r,k+1:k+r+1)];
   Y = -f*[X, 0] + Z;
   L(:,:,k) = [1, zeros(1,r); -f, eye(r)];
   R(k,k) = gamma;
   R(k,k+1:k+r) = X;
end
% last step
k=N-r;
gamma = Y(1,1);
   X = Y(1,2:r+1);
   f = [Y(2:r,1); A(k+r,k)]/gamma;
   % Z = [Y(2:r+1,2:r+2); A(k+r,k+1:k+r+2)];
   Z = [Y(2:r,2:r+1);A(k+r,k+1:k+r)];  
   Y = -f*X + Z;
   L(:,:,k) = [1, zeros(1,r); -f, eye(r)];
   R(k,k) = gamma;
   R(k,k+1:k+r) = X;

% 1.3
% Y is now Y_{N-r}
%Z = [Y; A(N,N-r+1:N)];
%[lastL,lastR] = myLU(Z);
[lastL,lastR] = myLU(Y);
L(1:r,1:r,N-r+1) = lastL;
R(N-r+1:N,N-r+1:N) = lastR;

%%%%%% Check (optional) %%%%%
% the factorization here is LA = R
% B = A;
% for k = 1:N-r
%     B(k:k+r,:) = L(1:r+1,1:r+1,k)*B(k:k+r,:);
% end
% B(N-r+1:N,:) = L(1:r,1:r,N-r+1)*B(N-r+1:N,:);
% disp(norm(R-B))
%%%%%% End check %%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Step 2: compute lower Green generators of inv(A)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Compute lower Green generators of L^{-1} 
% (Lemma 5.5)
pL = zeros(N-r,r);  % i-th row is pL(i)
dL = zeros(N-r,1);  % i-th entry is dL(i)
aL = zeros(r,r,N-1);% i-th matrix is aL(i)
qL = zeros(r,N-r);  % i-th column is qL(i)
% pL(N-r+1) is an rxr matrix and will be defined last
for k = 1:N-r
    pL(k,:) = L(1,1:r,k);
    dL(k) = L(1,r+1,k);
    aL(:,:,k) = L(2:r+1,1:r,k);
    qL(:,k) = L(2:r+1,r+1,k);
end
pLlast = L(1:r,1:r,N-r+1);

% 2.1: compute q and a
q = qL;   % N-r generators
a = aL;   % N-r generators

% 2.2: compute p
p = zeros(N-r,r);
Z = R(N-r+1:N,N-r+1:N);
Plast = Z\L(1:r,1:r,N-r+1);
t = Plast;
for k = N-r:-1:1
    p(k,:) = (pL(k,:)-R(k,k+1:k+r)*t*a(:,:,k))/R(k,k); % modified here
    t = [p(k,:); t(1:r-1,:)*a(:,:,k)];
end
% end of algorithm

end
    
    


